@@ -1,8 +1,10 @@
+use crate as scylla;
 use crate::cql_to_rust::FromCqlVal;
 use crate::frame::response::result::CqlValue;
 use crate::frame::value::Counter;
 use crate::frame::value::Value;
 use crate::frame::value::{Date, Time, Timestamp};
+use crate::macros::{FromUserType, IntoUserType};
 use crate::transport::session::IntoTypedRows;
 use crate::transport::session::Session;
 use crate::SessionBuilder;
@@ -203,20 +205,23 @@ async fn test_naive_date() {
 
     let tests = [
         // Basic test values
-        ("0000-1-1", Some(NaiveDate::from_ymd(0000, 1, 1))),
+        ("0000-01-01", Some(NaiveDate::from_ymd(0000, 1, 1))),
         ("1970-01-01", Some(NaiveDate::from_ymd(1970, 1, 1))),
         ("2020-03-07", Some(NaiveDate::from_ymd(2020, 3, 7))),
-        ("1337-4-5", Some(NaiveDate::from_ymd(1337, 4, 5))),
-        ("-1-12-31", Some(NaiveDate::from_ymd(-1, 12, 31))),
+        ("1337-04-05", Some(NaiveDate::from_ymd(1337, 4, 5))),
+        ("-0001-12-31", Some(NaiveDate::from_ymd(-1, 12, 31))),
         // min/max values allowed by NaiveDate
-        ("-262144-1-1", Some(min_naive_date)),
-        ("262143-12-31", Some(max_naive_date)),
+        ("-262144-01-01", Some(min_naive_date)),
+        // NOTICE: dropped for Cassandra 4 compatibility
+        //("262143-12-31", Some(max_naive_date)),
+
         // 1 less/more than min/max values allowed by NaiveDate
         ("-262145-12-31", None),
-        ("262144-1-1", None),
+        // NOTICE: dropped for Cassandra 4 compatibility
+        //("262144-01-01", None),
         // min/max values allowed by the database
         ("-5877641-06-23", None),
-        ("5881580-07-11", None),
+        //("5881580-07-11", None),
     ];
 
     for (date_text, date) in tests.iter() {
@@ -298,7 +303,8 @@ async fn test_date() {
         ("1969-12-02", Date(2_u32.pow(31) - 30)),
         ("1970-01-31", Date(2_u32.pow(31) + 30)),
         ("-5877641-06-23", Date(0)),
-        ("5881580-07-11", Date(u32::max_value())),
+        // NOTICE: dropped for Cassandra 4 compatibility
+        //("5881580-07-11", Date(u32::max_value())),
     ];
 
     for (date_text, date) in &tests {
@@ -428,13 +434,13 @@ async fn test_time() {
 async fn test_timestamp() {
     let session: Session = init_test("timestamp_tests", "timestamp").await;
 
-    let epoch_date = NaiveDate::from_ymd(1970, 1, 1);
+    //let epoch_date = NaiveDate::from_ymd(1970, 1, 1);
 
     //let before_epoch = NaiveDate::from_ymd(1333, 4, 30);
     //let before_epoch_offset = before_epoch.signed_duration_since(epoch_date);
 
-    let after_epoch = NaiveDate::from_ymd(2020, 3, 8);
-    let after_epoch_offset = after_epoch.signed_duration_since(epoch_date);
+    //let after_epoch = NaiveDate::from_ymd(2020, 3, 8);
+    //let after_epoch_offset = after_epoch.signed_duration_since(epoch_date);
 
     let tests = [
         ("0", Duration::milliseconds(0)),
@@ -446,8 +452,10 @@ async fn test_timestamp() {
             "-9223372036854775808",
             Duration::milliseconds(i64::min_value()),
         ),
-        ("1970-01-01", Duration::milliseconds(0)),
-        ("2020-03-08", after_epoch_offset),
+        // NOTICE: dropped for Cassandra 4 compatibility
+        //("1970-01-01", Duration::milliseconds(0)),
+        //("2020-03-08", after_epoch_offset),
+
         // Scylla rejects timestamps before 1970-01-01, but the specification says it shouldn't
         // https://github.com/apache/cassandra/blob/78b13cd0e7a33d45c2081bb135e860bbaca7cbe5/doc/native_protocol_v4.spec#L929
         // Scylla bug?
@@ -747,3 +755,152 @@ async fn test_blob() {
         assert_eq!(read_blob, *blob);
     }
 }
+
+#[tokio::test]
+async fn test_udt_after_schema_update() {
+    let table_name = "udt_tests";
+    let type_name = "usertype1";
+
+    let uri = env::var("SCYLLA_URI").unwrap_or_else(|_| "127.0.0.1:9042".to_string());
+
+    println!("Connecting to {} ...", uri);
+    let session: Session = SessionBuilder::new().known_node(uri).build().await.unwrap();
+
+    session
+        .query(
+            "CREATE KEYSPACE IF NOT EXISTS ks WITH REPLICATION = \
+            {'class' : 'SimpleStrategy', 'replication_factor' : 1}",
+            &[],
+        )
+        .await
+        .unwrap();
+
+    session
+        .query(format!("DROP TABLE IF EXISTS ks.{}", table_name), &[])
+        .await
+        .unwrap();
+
+    session
+        .query(format!("DROP TYPE IF EXISTS ks.{}", type_name), &[])
+        .await
+        .unwrap();
+
+    session
+        .query(
+            format!(
+                "CREATE TYPE IF NOT EXISTS ks.{} (first int, second boolean)",
+                type_name
+            ),
+            &[],
+        )
+        .await
+        .unwrap();
+
+    session
+        .query(
+            format!(
+                "CREATE TABLE IF NOT EXISTS ks.{} (id int PRIMARY KEY, val ks.{})",
+                table_name, type_name
+            ),
+            &[],
+        )
+        .await
+        .unwrap();
+
+    #[derive(IntoUserType, FromUserType, Debug, PartialEq)]
+    struct UdtV1 {
+        pub first: i32,
+        pub second: bool,
+    }
+
+    let v1 = UdtV1 {
+        first: 123,
+        second: true,
+    };
+
+    session
+        .query(
+            format!(
+                "INSERT INTO ks.{}(id,val) VALUES (0, {})",
+                table_name, "{first: 123, second: true}"
+            ),
+            &[],
+        )
+        .await
+        .unwrap();
+
+    let (read_udt,): (UdtV1,) = session
+        .query(
+            format!("SELECT val from ks.{} WHERE id = 0", table_name),
+            &[],
+        )
+        .await
+        .unwrap()
+        .rows
+        .unwrap()
+        .into_typed::<(UdtV1,)>()
+        .next()
+        .unwrap()
+        .unwrap();
+
+    assert_eq!(read_udt, v1);
+
+    session
+        .query(
+            format!("INSERT INTO ks.{}(id,val) VALUES (0, ?)", table_name),
+            &(&v1,),
+        )
+        .await
+        .unwrap();
+
+    let (read_udt,): (UdtV1,) = session
+        .query(
+            format!("SELECT val from ks.{} WHERE id = 0", table_name),
+            &[],
+        )
+        .await
+        .unwrap()
+        .rows
+        .unwrap()
+        .into_typed::<(UdtV1,)>()
+        .next()
+        .unwrap()
+        .unwrap();
+
+    assert_eq!(read_udt, v1);
+
+    session
+        .query(format!("ALTER TYPE ks.{} ADD third text;", type_name), &[])
+        .await
+        .unwrap();
+
+    #[derive(FromUserType, Debug, PartialEq)]
+    struct UdtV2 {
+        pub first: i32,
+        pub second: bool,
+        pub third: Option<String>,
+    }
+
+    let (read_udt,): (UdtV2,) = session
+        .query(
+            format!("SELECT val from ks.{} WHERE id = 0", table_name),
+            &[],
+        )
+        .await
+        .unwrap()
+        .rows
+        .unwrap()
+        .into_typed::<(UdtV2,)>()
+        .next()
+        .unwrap()
+        .unwrap();
+
+    assert_eq!(
+        read_udt,
+        UdtV2 {
+            first: 123,
+            second: true,
+            third: None,
+        }
+    );
+}
