@@ -9,7 +9,6 @@ use chrono::prelude::*;
 use chrono::Duration;
 use num_bigint::BigInt;
 use std::{
-    collections::BTreeMap,
     convert::{TryFrom, TryInto},
     net::IpAddr,
     result::Result as StdResult,
@@ -83,6 +82,7 @@ pub enum CqlValue {
     /// Can be converted to chrono::NaiveDate (-262145-1-1 to 262143-12-31) using as_date
     Date(u32),
     Double(f64),
+    Empty,
     Float(f32),
     Int(i32),
     BigInt(i64),
@@ -96,14 +96,17 @@ pub enum CqlValue {
     UserDefinedType {
         keyspace: String,
         type_name: String,
-        fields: BTreeMap<String, Option<CqlValue>>,
+        /// Order of `fields` vector must match the order of fields as defined in the UDT. The
+        /// driver does not check it by itself, so incorrect data will be written if the order is
+        /// wrong.
+        fields: Vec<(String, Option<CqlValue>)>,
     },
     SmallInt(i16),
     TinyInt(i8),
     /// Nanoseconds since midnight
     Time(Duration),
     Timeuuid(Uuid),
-    Tuple(Vec<CqlValue>),
+    Tuple(Vec<Option<CqlValue>>),
     Uuid(Uuid),
     Varint(BigInt),
 }
@@ -111,7 +114,7 @@ pub enum CqlValue {
 impl CqlValue {
     pub fn as_ascii(&self) -> Option<&String> {
         match self {
-            Self::Ascii(s) => Some(&s),
+            Self::Ascii(s) => Some(s),
             _ => None,
         }
     }
@@ -204,14 +207,14 @@ impl CqlValue {
 
     pub fn as_blob(&self) -> Option<&Vec<u8>> {
         match self {
-            Self::Blob(v) => Some(&v),
+            Self::Blob(v) => Some(v),
             _ => None,
         }
     }
 
     pub fn as_text(&self) -> Option<&String> {
         match self {
-            Self::Text(s) => Some(&s),
+            Self::Text(s) => Some(s),
             _ => None,
         }
     }
@@ -247,14 +250,14 @@ impl CqlValue {
 
     pub fn as_list(&self) -> Option<&Vec<CqlValue>> {
         match self {
-            Self::List(s) => Some(&s),
+            Self::List(s) => Some(s),
             _ => None,
         }
     }
 
     pub fn as_set(&self) -> Option<&Vec<CqlValue>> {
         match self {
-            Self::Set(s) => Some(&s),
+            Self::Set(s) => Some(s),
             _ => None,
         }
     }
@@ -293,7 +296,7 @@ impl CqlValue {
 #[derive(Debug, Clone)]
 pub struct ColumnSpec {
     pub table_spec: TableSpec,
-    name: String,
+    pub name: String,
     typ: ColumnType,
 }
 
@@ -301,7 +304,7 @@ pub struct ColumnSpec {
 pub struct ResultMetadata {
     col_count: usize,
     pub paging_state: Option<Bytes>,
-    col_specs: Vec<ColumnSpec>,
+    pub col_specs: Vec<ColumnSpec>,
 }
 
 #[derive(Debug, Clone)]
@@ -499,6 +502,16 @@ fn deser_prepared_metadata(buf: &mut &[u8]) -> StdResult<PreparedMetadata, Parse
 
 fn deser_cql_value(typ: &ColumnType, buf: &mut &[u8]) -> StdResult<CqlValue, ParseError> {
     use ColumnType::*;
+
+    if buf.is_empty() {
+        match typ {
+            Ascii | Blob | Text => {
+                // can't be empty
+            }
+            _ => return Ok(CqlValue::Empty),
+        }
+    }
+
     Ok(match typ {
         Ascii => {
             if !buf.is_ascii() {
@@ -703,15 +716,22 @@ fn deser_cql_value(typ: &ColumnType, buf: &mut &[u8]) -> StdResult<CqlValue, Par
             keyspace,
             field_types,
         } => {
-            let mut fields: BTreeMap<String, Option<CqlValue>> = BTreeMap::new();
+            let mut fields: Vec<(String, Option<CqlValue>)> = Vec::new();
 
             for (field_name, field_type) in field_types {
-                let mut field_value: Option<CqlValue> = None;
-                if let Some(mut field_val_bytes) = types::read_bytes_opt(buf)? {
-                    field_value = Some(deser_cql_value(&field_type, &mut field_val_bytes)?);
+                // If a field is added to a UDT and we read an old (frozen ?) version of it,
+                // the driver will fail to parse the whole UDT.
+                // This is why we break the parsing after we reach the end of the serialized UDT.
+                if buf.is_empty() {
+                    break;
                 }
 
-                fields.insert(field_name.clone(), field_value);
+                let mut field_value: Option<CqlValue> = None;
+                if let Some(mut field_val_bytes) = types::read_bytes_opt(buf)? {
+                    field_value = Some(deser_cql_value(field_type, &mut field_val_bytes)?);
+                }
+
+                fields.push((field_name.clone(), field_value));
             }
 
             CqlValue::UserDefinedType {
@@ -723,9 +743,12 @@ fn deser_cql_value(typ: &ColumnType, buf: &mut &[u8]) -> StdResult<CqlValue, Par
         Tuple(type_names) => {
             let mut res = Vec::with_capacity(type_names.len());
             for type_name in type_names {
-                let mut b = types::read_bytes(buf)?;
-                res.push(deser_cql_value(type_name, &mut b)?);
+                match types::read_bytes_opt(buf)? {
+                    Some(mut b) => res.push(Some(deser_cql_value(type_name, &mut b)?)),
+                    None => res.push(None),
+                };
             }
+
             CqlValue::Tuple(res)
         }
     })
@@ -1080,7 +1103,7 @@ mod tests {
         assert_eq!(date, CqlValue::Date(u32::max_value()));
 
         // Trying to parse a 0, 3 or 5 byte array fails
-        super::deser_cql_value(&ColumnType::Date, &mut [].as_ref()).unwrap_err();
+        super::deser_cql_value(&ColumnType::Date, &mut [].as_ref()).unwrap();
         super::deser_cql_value(&ColumnType::Date, &mut [1, 2, 3].as_ref()).unwrap_err();
         super::deser_cql_value(&ColumnType::Date, &mut [1, 2, 3, 4, 5].as_ref()).unwrap_err();
 
@@ -1198,6 +1221,49 @@ mod tests {
             );
         }
     }
+    #[test]
+    fn test_deserialize_empty_payload() {
+        for (test_type, res_cql) in [
+            (ColumnType::Ascii, CqlValue::Ascii("".to_owned())),
+            (ColumnType::Boolean, CqlValue::Empty),
+            (ColumnType::Blob, CqlValue::Blob(vec![])),
+            (ColumnType::Counter, CqlValue::Empty),
+            (ColumnType::Date, CqlValue::Empty),
+            (ColumnType::Decimal, CqlValue::Empty),
+            (ColumnType::Double, CqlValue::Empty),
+            (ColumnType::Float, CqlValue::Empty),
+            (ColumnType::Int, CqlValue::Empty),
+            (ColumnType::BigInt, CqlValue::Empty),
+            (ColumnType::Text, CqlValue::Text("".to_owned())),
+            (ColumnType::Timestamp, CqlValue::Empty),
+            (ColumnType::Inet, CqlValue::Empty),
+            (ColumnType::List(Box::new(ColumnType::Int)), CqlValue::Empty),
+            (
+                ColumnType::Map(Box::new(ColumnType::Int), Box::new(ColumnType::Int)),
+                CqlValue::Empty,
+            ),
+            (ColumnType::Set(Box::new(ColumnType::Int)), CqlValue::Empty),
+            (
+                ColumnType::UserDefinedType {
+                    type_name: "".to_owned(),
+                    keyspace: "".to_owned(),
+                    field_types: vec![],
+                },
+                CqlValue::Empty,
+            ),
+            (ColumnType::SmallInt, CqlValue::Empty),
+            (ColumnType::TinyInt, CqlValue::Empty),
+            (ColumnType::Time, CqlValue::Empty),
+            (ColumnType::Timeuuid, CqlValue::Empty),
+            (ColumnType::Tuple(vec![]), CqlValue::Empty),
+            (ColumnType::Uuid, CqlValue::Empty),
+            (ColumnType::Varint, CqlValue::Empty),
+        ] {
+            let cql_value: CqlValue = super::deser_cql_value(&test_type, &mut &[][..]).unwrap();
+
+            assert_eq!(cql_value, res_cql);
+        }
+    }
 
     #[test]
     fn test_timeuuid_deserialize() {
