@@ -30,7 +30,7 @@ use crate::batch::{Batch, BatchStatement};
 use crate::frame::{
     self,
     request::{self, batch, execute, query, register, Request},
-    response::{event::Event, result, Response, ResponseOpcode},
+    response::{event::Event, result, result::ColumnSpec, Response, ResponseOpcode},
     server_event_type::EventType,
     value::{BatchValues, ValueList},
     FrameParams, SerializedRequest,
@@ -38,6 +38,7 @@ use crate::frame::{
 use crate::query::Query;
 use crate::routing::ShardInfo;
 use crate::statement::prepared_statement::PreparedStatement;
+use crate::statement::Consistency;
 use crate::transport::session::IntoTypedRows;
 use crate::transport::Authenticator;
 use crate::transport::Authenticator::{
@@ -56,7 +57,7 @@ pub struct Connection {
     source_port: u16,
     shard_info: Option<ShardInfo>,
     config: ConnectionConfig,
-    is_shard_aware: bool,
+    shard_aware_port: Option<u16>,
 }
 
 type ResponseHandler = oneshot::Sender<Result<TaskResponse, QueryError>>;
@@ -90,6 +91,35 @@ pub struct QueryResult {
     pub tracing_id: Option<Uuid>,
     /// Paging state returned from the server
     pub paging_state: Option<Bytes>,
+    /// Column specification returned from the server
+    pub col_specs: Vec<ColumnSpec>,
+}
+
+impl QueryResult {
+    // Returns a column specification for a column with given name, or None if not found
+    pub fn get_column_spec<'a>(&'a self, name: &str) -> Option<(usize, &'a ColumnSpec)> {
+        self.col_specs
+            .iter()
+            .enumerate()
+            .find(|(_id, spec)| spec.name == name)
+    }
+
+    /// This function is used to merge results of multiple paged queries into one.  
+    /// other is the result of a new paged query.  
+    /// It is merged with current result kept in self.  
+    pub(crate) fn merge_with_next_page_res(&mut self, other: QueryResult) {
+        if let Some(other_rows) = other.rows {
+            match &mut self.rows {
+                Some(self_rows) => self_rows.extend(other_rows),
+                None => self.rows = Some(other_rows),
+            }
+        };
+
+        self.warnings.extend(other.warnings);
+        self.tracing_id = other.tracing_id;
+        self.paging_state = other.paging_state;
+        self.col_specs = other.col_specs;
+    }
 }
 
 /// Result of Session::batch(). Contains no rows, only some useful information.
@@ -101,11 +131,22 @@ pub struct BatchResult {
 }
 
 impl QueryResponse {
+    pub fn as_set_keyspace(&self) -> Option<&result::SetKeyspace> {
+        match &self.response {
+            Response::Result(result::Result::SetKeyspace(sk)) => Some(sk),
+            _ => None,
+        }
+    }
+
     pub fn into_query_result(self) -> Result<QueryResult, QueryError> {
-        let (rows, paging_state) = match self.response {
+        let (rows, paging_state, col_specs) = match self.response {
             Response::Error(err) => return Err(err.into()),
-            Response::Result(result::Result::Rows(rs)) => (Some(rs.rows), rs.metadata.paging_state),
-            Response::Result(_) => (None, None),
+            Response::Result(result::Result::Rows(rs)) => (
+                Some(rs.rows),
+                rs.metadata.paging_state,
+                rs.metadata.col_specs,
+            ),
+            Response::Result(_) => (None, None, vec![]),
             _ => {
                 return Err(QueryError::ProtocolError(
                     "Unexpected server response, expected Result or Error",
@@ -118,6 +159,7 @@ impl QueryResponse {
             warnings: self.warnings,
             tracing_id: self.tracing_id,
             paging_state,
+            col_specs,
         })
     }
 }
@@ -133,6 +175,7 @@ pub struct ConnectionConfig {
     pub connect_timeout: std::time::Duration,
     // should be Some only in control connections,
     pub event_sender: Option<mpsc::Sender<Event>>,
+    pub default_consistency: Consistency,
     /*
     These configuration options will be added in the future:
 
@@ -140,8 +183,6 @@ pub struct ConnectionConfig {
 
     pub load_balancing: Option<String>,
     pub retry_policy: Option<String>,
-
-    pub default_consistency: Option<String>,
     */
 }
 
@@ -156,10 +197,23 @@ impl Default for ConnectionConfig {
             auth_username: None,
             auth_password: None,
             connect_timeout: std::time::Duration::from_secs(5),
+            default_consistency: Default::default(),
         }
     }
 }
 
+impl ConnectionConfig {
+    #[cfg(feature = "ssl")]
+    pub fn is_ssl(&self) -> bool {
+        self.ssl_context.is_some()
+    }
+
+    #[cfg(not(feature = "ssl"))]
+    pub fn is_ssl(&self) -> bool {
+        false
+    }
+}
+
 // Used to listen for fatal error in connection
 pub type ErrorReceiver = tokio::sync::oneshot::Receiver<QueryError>;
 
@@ -201,7 +255,7 @@ impl Connection {
             connect_address: addr,
             shard_info: None,
             config,
-            is_shard_aware: false,
+            shard_aware_port: None,
         };
 
         Ok((connection, error_receiver))
@@ -225,7 +279,7 @@ impl Connection {
         let query_response = self
             .send_request(
                 &request::Prepare {
-                    query: &query.get_contents(),
+                    query: &query.contents,
                 },
                 true,
                 query.config.tracing,
@@ -237,7 +291,7 @@ impl Connection {
             Response::Result(result::Result::Prepared(p)) => PreparedStatement::new(
                 p.id,
                 p.prepared_metadata,
-                query.get_contents().to_owned(),
+                query.contents.clone(),
                 query.get_page_size(),
             ),
             _ => {
@@ -300,13 +354,16 @@ impl Connection {
         let serialized_values = values.serialized()?;
 
         let query_frame = query::Query {
-            contents: query.get_contents().to_owned(),
+            contents: &query.contents,
             parameters: query::QueryParameters {
-                consistency: query.get_consistency(),
+                consistency: query
+                    .config
+                    .determine_consistency(self.config.default_consistency),
                 serial_consistency: query.get_serial_consistency(),
                 values: &serialized_values,
                 page_size: query.get_page_size(),
                 paging_state,
+                timestamp: query.get_timestamp(),
             },
         };
 
@@ -314,6 +371,45 @@ impl Connection {
             .await
     }
 
+    /// Performs query_single_page multiple times to query all available pages
+    pub async fn query_all(
+        &self,
+        query: &Query,
+        values: impl ValueList,
+    ) -> Result<QueryResult, QueryError> {
+        if query.get_page_size().is_none() {
+            // Page size should be set when someone wants to use paging
+            // This is an internal function, we can use ProtocolError even if it's not technically desigen for this
+            return Err(QueryError::ProtocolError(
+                "Called Connection::query_all without page size set!",
+            ));
+        }
+
+        let mut final_result = QueryResult::default();
+
+        let serialized_values = values.serialized()?;
+        let mut paging_state: Option<Bytes> = None;
+
+        loop {
+            // Send next paged query
+            let mut cur_result: QueryResult = self
+                .query(query, &serialized_values, paging_state)
+                .await?
+                .into_query_result()?;
+
+            // Set paging_state for the next query
+            paging_state = cur_result.paging_state.take();
+
+            // Add current query results to the final_result
+            final_result.merge_with_next_page_res(cur_result);
+
+            if paging_state.is_none() {
+                // No more pages to query, we can return the final result
+                return Ok(final_result);
+            }
+        }
+    }
+
     pub async fn execute_single_page(
         &self,
         prepared_statement: &PreparedStatement,
@@ -336,10 +432,13 @@ impl Connection {
         let execute_frame = execute::Execute {
             id: prepared_statement.get_id().to_owned(),
             parameters: query::QueryParameters {
-                consistency: prepared_statement.get_consistency(),
+                consistency: prepared_statement
+                    .config
+                    .determine_consistency(self.config.default_consistency),
                 serial_consistency: prepared_statement.get_serial_consistency(),
                 values: &serialized_values,
                 page_size: prepared_statement.get_page_size(),
+                timestamp: prepared_statement.get_timestamp(),
                 paging_state,
             },
         };
@@ -370,12 +469,49 @@ impl Connection {
         Ok(query_response)
     }
 
+    /// Performs execute_single_page multiple times to fetch all available pages
+    pub async fn execute_all(
+        &self,
+        prepared_statement: &PreparedStatement,
+        values: impl ValueList,
+    ) -> Result<QueryResult, QueryError> {
+        if prepared_statement.get_page_size().is_none() {
+            // This is an internal function, we can use ProtocolError even if it's not technically desigen for this
+            return Err(QueryError::ProtocolError(
+                "Called Connection::execute_all without page size set!",
+            ));
+        }
+
+        let mut final_result = QueryResult::default();
+
+        let serialized_values = values.serialized()?;
+        let mut paging_state: Option<Bytes> = None;
+
+        loop {
+            // Send next paged query
+            let mut cur_result: QueryResult = self
+                .execute_single_page(prepared_statement, &serialized_values, paging_state)
+                .await?;
+
+            // Set paging_state for the next query
+            paging_state = cur_result.paging_state.take();
+
+            // Add current query results to the final_result
+            final_result.merge_with_next_page_res(cur_result);
+
+            if paging_state.is_none() {
+                // No more pages to query, we can return the final result
+                return Ok(final_result);
+            }
+        }
+    }
+
     pub async fn batch(
         &self,
         batch: &Batch,
         values: impl BatchValues,
     ) -> Result<BatchResult, QueryError> {
-        let statements_count = batch.get_statements().len();
+        let statements_count = batch.statements.len();
         if statements_count != values.len() {
             return Err(QueryError::BadQuery(BadQuery::ValueLenMismatch(
                 values.len(),
@@ -383,10 +519,8 @@ impl Connection {
             )));
         }
 
-        let statements_iter = batch.get_statements().iter().map(|s| match s {
-            BatchStatement::Query(q) => batch::BatchStatement::Query {
-                text: q.get_contents(),
-            },
+        let statements_iter = batch.statements.iter().map(|s| match s {
+            BatchStatement::Query(q) => batch::BatchStatement::Query { text: &q.contents },
             BatchStatement::PreparedStatement(s) => {
                 batch::BatchStatement::Prepared { id: s.get_id() }
             }
@@ -397,8 +531,11 @@ impl Connection {
             statements_count,
             values,
             batch_type: batch.get_type(),
-            consistency: batch.get_consistency(),
+            consistency: batch
+                .config
+                .determine_consistency(self.config.default_consistency),
             serial_consistency: batch.get_serial_consistency(),
+            timestamp: batch.get_timestamp(),
         };
 
         let query_response = self
@@ -745,7 +882,11 @@ impl Connection {
     /// Are we connected to Scylla's shard aware port?
     // TODO: couple this with shard_info?
     pub fn get_is_shard_aware(&self) -> bool {
-        self.is_shard_aware
+        Some(self.connect_address.port()) == self.shard_aware_port
+    }
+
+    pub fn get_shard_aware_port(&self) -> Option<u16> {
+        self.shard_aware_port
     }
 
     pub fn get_source_port(&self) -> u16 {
@@ -756,8 +897,8 @@ impl Connection {
         self.shard_info = shard_info
     }
 
-    fn set_is_shard_aware(&mut self, is_shard_aware: bool) {
-        self.is_shard_aware = is_shard_aware;
+    fn set_shard_aware_port(&mut self, shard_aware_port: Option<u16>) {
+        self.shard_aware_port = shard_aware_port;
     }
 
     pub fn get_connect_address(&self) -> SocketAddr {
@@ -791,6 +932,11 @@ pub async fn open_named_connection(
 
     let options_result = connection.get_options().await?;
 
+    let shard_aware_port_key = match config.is_ssl() {
+        true => "SCYLLA_SHARD_AWARE_PORT_SSL",
+        false => "SCYLLA_SHARD_AWARE_PORT",
+    };
+
     let (shard_info, supported_compression, shard_aware_port) = match options_result {
         Response::Supported(mut supported) => {
             let shard_info = ShardInfo::try_from(&supported.options).ok();
@@ -800,7 +946,7 @@ pub async fn open_named_connection(
                 .unwrap_or_else(Vec::new);
             let shard_aware_port = supported
                 .options
-                .remove("SCYLLA_SHARD_AWARE_PORT")
+                .remove(shard_aware_port_key)
                 .unwrap_or_else(Vec::new)
                 .into_iter()
                 .next()
@@ -810,7 +956,7 @@ pub async fn open_named_connection(
         _ => (None, Vec::new(), None),
     };
     connection.set_shard_info(shard_info);
-    connection.set_is_shard_aware(Some(addr.port()) == shard_aware_port);
+    connection.set_shard_aware_port(shard_aware_port);
 
     let mut options = HashMap::new();
     options.insert("CQL_VERSION".to_string(), "4.0.0".to_string()); // FIXME: hardcoded values
@@ -1034,3 +1180,126 @@ impl VerifiedKeyspaceName {
         Ok(())
     }
 }
+
+#[cfg(test)]
+mod tests {
+    use super::super::errors::QueryError;
+    use super::ConnectionConfig;
+    use crate::query::Query;
+    use crate::IntoTypedRows;
+    use std::net::SocketAddr;
+
+    // Just like resolve_hostname in session.rs
+    async fn resolve_hostname(hostname: &str) -> SocketAddr {
+        match tokio::net::lookup_host(hostname).await {
+            Ok(mut addrs) => addrs.next().unwrap(),
+            Err(_) => {
+                tokio::net::lookup_host((hostname, 9042)) // Port might not be specified, try default
+                    .await
+                    .unwrap()
+                    .next()
+                    .unwrap()
+            }
+        }
+    }
+
+    /// Tests for Connection::query_all and Connection::execute_all
+    /// 1. SELECT from an empty table.
+    /// 2. Create table and insert ints 0..100.
+    ///    Then use query_all and execute_all with page_size set to 7 to select all 100 rows.
+    /// 3. INSERT query_all should have None in result rows.
+    /// 4. Calling query_all with a Query that doesn't have page_size set should result in an error.
+    #[tokio::test]
+    async fn connection_query_all_execute_all_test() {
+        let uri = std::env::var("SCYLLA_URI").unwrap_or_else(|_| "127.0.0.1:9042".to_string());
+        let addr: SocketAddr = resolve_hostname(&uri).await;
+
+        let (connection, _) = super::open_connection(addr, None, ConnectionConfig::default())
+            .await
+            .unwrap();
+
+        connection.query_single_page("CREATE KEYSPACE IF NOT EXISTS ks WITH REPLICATION = {'class' : 'SimpleStrategy', 'replication_factor' : 1}", &[]).await.unwrap();
+        connection
+            .query_single_page("DROP TABLE IF EXISTS ks.connection_query_all_tab", &[])
+            .await
+            .unwrap();
+        connection
+            .query_single_page(
+                "CREATE TABLE IF NOT EXISTS ks.connection_query_all_tab (p int primary key)",
+                &[],
+            )
+            .await
+            .unwrap();
+
+        // 1. SELECT from an empty table returns query result where rows are Some(Vec::new())
+        let select_query =
+            Query::new("SELECT p FROM ks.connection_query_all_tab").with_page_size(7);
+        let empty_res = connection.query_all(&select_query, &[]).await.unwrap();
+        assert!(empty_res.rows.unwrap().is_empty());
+
+        let mut prepared_select = connection.prepare(&select_query).await.unwrap();
+        prepared_select.set_page_size(7);
+        let empty_res_prepared = connection.execute_all(&prepared_select, &[]).await.unwrap();
+        assert!(empty_res_prepared.rows.unwrap().is_empty());
+
+        // 2. Insert 100 and select using query_all with page_size 7
+        let values: Vec<i32> = (0..100).collect();
+        let mut insert_futures = Vec::new();
+        let insert_query =
+            Query::new("INSERT INTO ks.connection_query_all_tab (p) VALUES (?)").with_page_size(7);
+        for v in &values {
+            insert_futures.push(connection.query_single_page(insert_query.clone(), (v,)));
+        }
+
+        futures::future::try_join_all(insert_futures).await.unwrap();
+
+        let mut results: Vec<i32> = connection
+            .query_all(&select_query, &[])
+            .await
+            .unwrap()
+            .rows
+            .unwrap()
+            .into_typed::<(i32,)>()
+            .map(|r| r.unwrap().0)
+            .collect();
+        results.sort_unstable(); // Clippy recommended to use sort_unstable instead of sort()
+        assert_eq!(results, values);
+
+        let mut results2: Vec<i32> = connection
+            .execute_all(&prepared_select, &[])
+            .await
+            .unwrap()
+            .rows
+            .unwrap()
+            .into_typed::<(i32,)>()
+            .map(|r| r.unwrap().0)
+            .collect();
+        results2.sort_unstable();
+        assert_eq!(results2, values);
+
+        // 3. INSERT query_all should have None in result rows.
+        let insert_res1 = connection.query_all(&insert_query, (0,)).await.unwrap();
+        assert!(insert_res1.rows.is_none());
+
+        let prepared_insert = connection.prepare(&insert_query).await.unwrap();
+        let insert_res2 = connection
+            .execute_all(&prepared_insert, (0,))
+            .await
+            .unwrap();
+        assert!(insert_res2.rows.is_none(),);
+
+        // 4. Calling query_all with a Query that doesn't have page_size set should result in an error.
+        let no_page_size_query = Query::new("SELECT p FROM ks.connection_query_all_tab");
+        let no_page_res = connection.query_all(&no_page_size_query, &[]).await;
+        assert!(matches!(no_page_res, Err(QueryError::ProtocolError(_))));
+
+        let prepared_no_page_size_query = connection.prepare(&no_page_size_query).await.unwrap();
+        let prepared_no_page_res = connection
+            .execute_all(&prepared_no_page_size_query, &[])
+            .await;
+        assert!(matches!(
+            prepared_no_page_res,
+            Err(QueryError::ProtocolError(_))
+        ));
+    }
+}
