@@ -3,24 +3,27 @@
 
 use bytes::Bytes;
 use futures::future::join_all;
+use futures::future::try_join_all;
 use std::future::Future;
 use std::net::SocketAddr;
 use std::sync::Arc;
 use std::time::Duration;
 use tokio::net::lookup_host;
 use tokio::time::timeout;
-use tracing::{info, warn};
+use tracing::debug;
 use uuid::Uuid;
 
+use super::connection::QueryResponse;
 use super::errors::{BadQuery, NewSessionError, QueryError};
 use crate::frame::response::cql_to_rust::FromRowError;
-use crate::frame::response::{result, Response};
+use crate::frame::response::result;
 use crate::frame::value::{BatchValues, SerializedValues, ValueList};
 use crate::prepared_statement::{PartitionKeyError, PreparedStatement};
 use crate::query::Query;
 use crate::routing::{murmur3_token, Token};
-use crate::statement::Consistency;
+use crate::statement::{Consistency, SerialConsistency};
 use crate::tracing::{GetTracingConfig, TracingEvent, TracingInfo};
+use crate::transport::connection_pool::PoolConfig;
 use crate::transport::{
     cluster::Cluster,
     connection::{BatchResult, Connection, ConnectionConfig, QueryResult, VerifiedKeyspaceName},
@@ -35,6 +38,9 @@ use crate::transport::{
 use crate::{batch::Batch, statement::StatementConfig};
 use crate::{cql_to_rust::FromRow, transport::speculative_execution};
 
+pub use crate::transport::connection_pool::PoolSize;
+
+use crate::transport::iterator::PreparedIteratorConfig;
 #[cfg(feature = "ssl")]
 use openssl::ssl::SslContext;
 
@@ -45,8 +51,8 @@ pub struct Session {
     schema_agreement_interval: Duration,
     retry_policy: Box<dyn RetryPolicy>,
     speculative_execution_policy: Option<Arc<dyn SpeculativeExecutionPolicy>>,
-
     metrics: Arc<Metrics>,
+    default_consistency: Consistency,
 }
 
 /// Configuration options for [`Session`].
@@ -82,13 +88,21 @@ pub struct SessionConfig {
 
     pub schema_agreement_interval: Duration,
     pub connect_timeout: std::time::Duration,
+
+    /// Size of the per-node connection pool, i.e. how many connections the driver should keep to each node.
+    /// The default is `PerShard(1)`, which is the recommended setting for Scylla clusters.
+    pub connection_pool_size: PoolSize,
+
+    /// If true, prevents the driver from connecting to the shard-aware port, even if the node supports it.
+    /// Generally, this options is best left as default (false).
+    pub disallow_shard_aware_port: bool,
+
+    pub default_consistency: Consistency,
     /*
     These configuration options will be added in the future:
 
 
     pub tcp_keepalive: bool,
-
-    pub default_consistency: Option<String>,
     */
 }
 
@@ -126,6 +140,9 @@ impl SessionConfig {
             auth_username: None,
             auth_password: None,
             connect_timeout: std::time::Duration::from_secs(5),
+            connection_pool_size: Default::default(),
+            disallow_shard_aware_port: false,
+            default_consistency: Consistency::Quorum,
         }
     }
 
@@ -187,6 +204,15 @@ impl SessionConfig {
         }
     }
 
+    /// Creates a PoolConfig which can be used to create NodeConnectionPools
+    fn get_pool_config(&self) -> PoolConfig {
+        PoolConfig {
+            connection_config: self.get_connection_config(),
+            pool_size: self.connection_pool_size.clone(),
+            can_use_shard_aware_port: !self.disallow_shard_aware_port,
+        }
+    }
+
     /// Makes a config that should be used in Connection
     fn get_connection_config(&self) -> ConnectionConfig {
         ConnectionConfig {
@@ -197,7 +223,8 @@ impl SessionConfig {
             auth_username: self.auth_username.to_owned(),
             auth_password: self.auth_password.to_owned(),
             connect_timeout: self.connect_timeout,
-            ..Default::default()
+            event_sender: None,
+            default_consistency: self.default_consistency,
         }
     }
 }
@@ -280,7 +307,7 @@ impl Session {
 
         for node in &config.known_nodes {
             match node {
-                KnownNode::Hostname(hostname) => to_resolve.push(&hostname),
+                KnownNode::Hostname(hostname) => to_resolve.push(hostname),
                 KnownNode::Address(address) => node_addresses.push(*address),
             };
         }
@@ -290,38 +317,7 @@ impl Session {
 
         node_addresses.extend(resolved);
 
-        let use_ssl = match () {
-            #[cfg(not(feature = "ssl"))]
-            () => false,
-            #[cfg(feature = "ssl")]
-            () => config.ssl_context.is_some(),
-        };
-
-        let mut shard_aware_addresses: Vec<SocketAddr> = vec![];
-        if let Some(shard_aware_port) =
-            Self::get_shard_aware_port(node_addresses[0], config.get_connection_config(), use_ssl)
-                .await
-        {
-            info!("Shard-aware port detected: {}", shard_aware_port);
-            shard_aware_addresses = (&node_addresses)
-                .iter()
-                .map(|addr| SocketAddr::new(addr.ip(), shard_aware_port))
-                .collect();
-        }
-
-        // Start the session
-        let cluster = if !shard_aware_addresses.is_empty() {
-            match Cluster::new(&shard_aware_addresses, config.get_connection_config()).await {
-                Ok(clust) => clust,
-                Err(e) => {
-                    warn!("Unable to establish connections at detected shard-aware port, falling back to default ports: {}", e);
-                    Cluster::new(&node_addresses, config.get_connection_config()).await?
-                }
-            }
-        } else {
-            info!("Shard-aware ports not available, falling back to default ports");
-            Cluster::new(&node_addresses, config.get_connection_config()).await?
-        };
+        let cluster = Cluster::new(&node_addresses, config.get_pool_config()).await?;
 
         let session = Session {
             cluster,
@@ -330,6 +326,7 @@ impl Session {
             schema_agreement_interval: config.schema_agreement_interval,
             speculative_execution_policy: config.speculative_execution_policy,
             metrics: Arc::new(Metrics::new()),
+            default_consistency: config.default_consistency,
         };
 
         if let Some(keyspace_name) = config.used_keyspace {
@@ -341,29 +338,6 @@ impl Session {
         Ok(session)
     }
 
-    async fn get_shard_aware_port(
-        addr: SocketAddr,
-        config: ConnectionConfig,
-        use_ssl: bool,
-    ) -> Option<u16> {
-        let (probe, _) = Connection::new(addr, None, config).await.ok()?;
-        let options_result = probe.get_options().await.ok()?;
-        let options_key = if use_ssl {
-            "SCYLLA_SHARD_AWARE_PORT_SSL"
-        } else {
-            "SCYLLA_SHARD_AWARE_PORT"
-        };
-        match options_result {
-            Response::Supported(mut supported) => supported
-                .options
-                .remove(options_key)
-                .unwrap_or_else(Vec::new)
-                .get(0)
-                .and_then(|p| p.parse::<u16>().ok()),
-            _ => None,
-        }
-    }
-
     /// Sends a query to the database and receives a response.  
     /// Returns only a single page of results, to receive multiple pages use [query_iter](Session::query_iter)
     ///
@@ -431,39 +405,46 @@ impl Session {
         paging_state: Option<Bytes>,
     ) -> Result<QueryResult, QueryError> {
         let query: Query = query.into();
-        let query_text: &str = query.get_contents();
-        let serialized_values = values.serialized();
+        let serialized_values = values.serialized()?;
 
-        // In case the user tried doing session.query("use keyspace ks") run session::use_keyspace
-        if query_is_setting_keyspace(query_text) {
-            warn!("Raw USE KEYSPACE queries are experimental, please use session::use_keyspace instead");
+        let response = self
+            .run_query(
+                Statement::default(),
+                &query.config,
+                |node: Arc<Node>| async move { node.random_connection().await },
+                |connection: Arc<Connection>| {
+                    // Needed to avoid moving query and values into async move block
+                    let query_ref = &query;
+                    let values_ref = &serialized_values;
+                    let paging_state_ref = &paging_state;
 
-            let keyspace_name = &query_text["use ".len()..].trim_end_matches(';').trim();
-            let case_sensitive = keyspace_name.starts_with('"');
-            let keyspace_name = keyspace_name.trim_matches('"');
+                    async move {
+                        connection
+                            .query(query_ref, values_ref, paging_state_ref.clone())
+                            .await
+                    }
+                },
+            )
+            .await?;
+        self.handle_set_keyspace_response(&response).await?;
 
-            return self
-                .use_keyspace(keyspace_name, case_sensitive)
-                .await
-                .map(|_| QueryResult::default());
+        response.into_query_result()
+    }
+
+    async fn handle_set_keyspace_response(
+        &self,
+        response: &QueryResponse,
+    ) -> Result<(), QueryError> {
+        if let Some(set_keyspace) = response.as_set_keyspace() {
+            debug!(
+                "Detected USE KEYSPACE query, setting session's keyspace to {}",
+                set_keyspace.keyspace_name
+            );
+            self.use_keyspace(set_keyspace.keyspace_name.clone(), true)
+                .await?;
         }
 
-        // Needed to avoid moving query and values into async move block
-        let query_ref: &Query = &query;
-        let values_ref = &serialized_values;
-        let paging_state_ref = &paging_state;
-
-        self.run_query(
-            Statement::default(),
-            &query.config,
-            |node: Arc<Node>| async move { node.random_connection().await },
-            |connection: Arc<Connection>| async move {
-                connection
-                    .query_single_page_by_ref(query_ref, values_ref, paging_state_ref.clone())
-                    .await
-            },
-        )
-        .await
+        Ok(())
     }
 
     /// Run a simple query with paging  
@@ -515,6 +496,7 @@ impl Session {
         Ok(RowIterator::new_for_query(
             query,
             serialized_values.into_owned(),
+            self.default_consistency,
             retry_session,
             self.load_balancer.clone(),
             self.cluster.get_data(),
@@ -558,7 +540,7 @@ impl Session {
     /// # }
     /// ```
     pub async fn prepare(&self, query: impl Into<Query>) -> Result<PreparedStatement, QueryError> {
-        let query: Query = query.into();
+        let query = query.into();
 
         let connections = self.cluster.get_working_connections().await?;
 
@@ -668,17 +650,21 @@ impl Session {
             keyspace: prepared.get_keyspace_name(),
         };
 
-        self.run_query(
-            statement_info,
-            &prepared.config,
-            |node: Arc<Node>| async move { node.connection_for_token(token).await },
-            |connection: Arc<Connection>| async move {
-                connection
-                    .execute_single_page(prepared, values_ref, paging_state_ref.clone())
-                    .await
-            },
-        )
-        .await
+        let response = self
+            .run_query(
+                statement_info,
+                &prepared.config,
+                |node: Arc<Node>| async move { node.connection_for_token(token).await },
+                |connection: Arc<Connection>| async move {
+                    connection
+                        .execute(prepared, values_ref, paging_state_ref.clone())
+                        .await
+                },
+            )
+            .await?;
+        self.handle_set_keyspace_response(&response).await?;
+
+        response.into_query_result()
     }
 
     /// Run a prepared query with paging  
@@ -727,7 +713,7 @@ impl Session {
         prepared: impl Into<PreparedStatement>,
         values: impl ValueList,
     ) -> Result<RowIterator, QueryError> {
-        let prepared: PreparedStatement = prepared.into();
+        let prepared = prepared.into();
         let serialized_values = values.serialized()?;
 
         let token = calculate_token(&prepared, &serialized_values)?;
@@ -738,13 +724,16 @@ impl Session {
         };
 
         Ok(RowIterator::new_for_prepared_statement(
-            prepared,
-            serialized_values.into_owned(),
-            token,
-            retry_session,
-            self.load_balancer.clone(),
-            self.cluster.get_data(),
-            self.metrics.clone(),
+            PreparedIteratorConfig {
+                prepared,
+                values: serialized_values.into_owned(),
+                default_consistency: self.default_consistency,
+                token,
+                retry_session,
+                load_balancer: self.load_balancer.clone(),
+                cluster_data: self.cluster.get_data(),
+                metrics: self.metrics.clone(),
+            },
         ))
     }
 
@@ -892,7 +881,7 @@ impl Session {
         // config.attempts is NonZeroU32 so at least one attempt will be made
         for _ in 0..config.attempts.get() {
             let current_try: Option<TracingInfo> = self
-                .try_getting_tracing_info(tracing_id, config.consistency)
+                .try_getting_tracing_info(tracing_id, Some(config.consistency))
                 .await?;
 
             match current_try {
@@ -915,18 +904,17 @@ impl Session {
     async fn try_getting_tracing_info(
         &self,
         tracing_id: &Uuid,
-        consistency: Consistency,
+        consistency: Option<Consistency>,
     ) -> Result<Option<TracingInfo>, QueryError> {
         // Query system_traces.sessions for TracingInfo
-        let mut traces_session_query =
-            Query::new(crate::tracing::TRACES_SESSION_QUERY_STR.to_string());
+        let mut traces_session_query = Query::new(crate::tracing::TRACES_SESSION_QUERY_STR);
         traces_session_query.config.consistency = consistency;
+        traces_session_query.set_page_size(1024);
 
         // Query system_traces.events for TracingEvents
-        let mut traces_events_query =
-            Query::new(crate::tracing::TRACES_EVENTS_QUERY_STR.to_string());
-        traces_events_query.config.consistency = Consistency::One;
+        let mut traces_events_query = Query::new(crate::tracing::TRACES_EVENTS_QUERY_STR);
         traces_events_query.config.consistency = consistency;
+        traces_events_query.set_page_size(1024);
 
         let (traces_session_res, traces_events_res) = tokio::try_join!(
             self.query(traces_session_query, (tracing_id,)),
@@ -1081,7 +1069,7 @@ impl Session {
         &self,
         query_plan: impl Iterator<Item = Arc<Node>>,
         is_idempotent: bool,
-        consistency: Consistency,
+        consistency: Option<Consistency>,
         mut retry_session: Box<dyn RetrySession>,
         choose_connection: impl Fn(Arc<Node>) -> ConnFut,
         do_query: impl Fn(Arc<Connection>) -> QueryFut,
@@ -1125,7 +1113,7 @@ impl Session {
                 let query_info = QueryInfo {
                     error: last_error.as_ref().unwrap(),
                     is_idempotent,
-                    consistency,
+                    consistency: consistency.unwrap_or(self.default_consistency),
                 };
 
                 match retry_session.decide_should_retry(query_info) {
@@ -1171,7 +1159,7 @@ impl Session {
         let info = Statement::default();
         let config = StatementConfig {
             is_idempotent: true,
-            serial_consistency: Some(Consistency::One),
+            serial_consistency: Some(SerialConsistency::LocalSerial),
             ..Default::default()
         };
 
@@ -1188,11 +1176,8 @@ impl Session {
         let connections = self.cluster.get_working_connections().await?;
 
         let handles = connections.iter().map(|c| c.fetch_schema_version());
-        let results = join_all(handles).await;
-        let versions: Vec<Uuid> = results
-            .iter()
-            .cloned()
-            .collect::<Result<Vec<_>, QueryError>>()?;
+        let versions = try_join_all(handles).await?;
+
         let local_version: Uuid = versions[0];
         let in_agreement = versions.into_iter().all(|v| v == local_version);
         Ok(in_agreement)
@@ -1206,17 +1191,6 @@ impl Session {
     }
 }
 
-/// Checks if a query sets a keyspace
-fn query_is_setting_keyspace(query: &str) -> bool {
-    let query_bytes = query.as_bytes();
-
-    if query_bytes.len() < 4 {
-        return false;
-    }
-
-    query_bytes[0..=3].eq_ignore_ascii_case("use ".as_bytes())
-}
-
 fn calculate_token(
     stmt: &PreparedStatement,
     values: &SerializedValues,
@@ -1263,18 +1237,3 @@ async fn resolve_hostname(hostname: &str) -> Result<SocketAddr, NewSessionError>
 
     ret.ok_or(failed_err)
 }
-
-#[cfg(test)]
-mod tests {
-    use super::*;
-
-    #[test]
-    fn test_query_is_setting_keyspace() {
-        assert!(query_is_setting_keyspace("use some_keyspace"));
-        assert!(query_is_setting_keyspace("UsE anotherKeySpace;"));
-        assert!(query_is_setting_keyspace("USE SCREAMINGKEYSPACE"));
-        assert!(!query_is_setting_keyspace("select * from users;"));
-        assert!(!query_is_setting_keyspace("us"));
-        assert!(!query_is_setting_keyspace(""));
-    }
-}
