@@ -1,3 +1,4 @@
+use crate as scylla;
 use crate::batch::Batch;
 use crate::frame::response::result::Row;
 use crate::frame::value::ValueList;
@@ -44,19 +45,24 @@ async fn test_unprepared_statement() {
         .await
         .unwrap();
 
-    let rs = session
+    let query_result = session
         .query("SELECT a, b, c FROM ks.t", &[])
         .await
-        .unwrap()
-        .rows
         .unwrap();
 
+    let (a_idx, _) = query_result.get_column_spec("a").unwrap();
+    let (b_idx, _) = query_result.get_column_spec("b").unwrap();
+    let (c_idx, _) = query_result.get_column_spec("c").unwrap();
+    assert!(query_result.get_column_spec("d").is_none());
+
+    let rs = query_result.rows.unwrap();
+
     let mut results: Vec<(i32, i32, &String)> = rs
         .iter()
         .map(|r| {
-            let a = r.columns[0].as_ref().unwrap().as_int().unwrap();
-            let b = r.columns[1].as_ref().unwrap().as_int().unwrap();
-            let c = r.columns[2].as_ref().unwrap().as_text().unwrap();
+            let a = r.columns[a_idx].as_ref().unwrap().as_int().unwrap();
+            let b = r.columns[b_idx].as_ref().unwrap().as_int().unwrap();
+            let c = r.columns[c_idx].as_ref().unwrap().as_text().unwrap();
             (a, b, c)
         })
         .collect();
@@ -70,7 +76,7 @@ async fn test_unprepared_statement() {
         ]
     );
     let mut results_from_manual_paging: Vec<Row> = vec![];
-    let query = Query::new("SELECT a, b, c FROM ks.t".to_owned()).with_page_size(1);
+    let query = Query::new("SELECT a, b, c FROM ks.t").with_page_size(1);
     let mut paging_state: Option<Bytes> = None;
     let mut watchdog = 0;
     loop {
@@ -190,7 +196,7 @@ async fn test_prepared_statement() {
         assert_eq!((a, b, c), (17, 16, &String::from("I'm prepared!!!")));
 
         let mut results_from_manual_paging: Vec<Row> = vec![];
-        let query = Query::new("SELECT a, b, c FROM ks.t2".to_owned()).with_page_size(1);
+        let query = Query::new("SELECT a, b, c FROM ks.t2").with_page_size(1);
         let prepared_paged = session.prepare(query).await.unwrap();
         let mut paging_state: Option<Bytes> = None;
         let mut watchdog = 0;
@@ -224,6 +230,43 @@ async fn test_prepared_statement() {
         assert!(e.is_none());
         assert_eq!((a, b, c, d), (17, 16, &String::from("I'm prepared!!!"), 7))
     }
+    // Check that ValueList macro works
+    {
+        #[derive(scylla::ValueList, scylla::FromRow, PartialEq, Debug, Clone)]
+        struct ComplexPk {
+            a: i32,
+            b: i32,
+            c: Option<String>,
+            d: i32,
+            e: i32,
+        }
+        let input: ComplexPk = ComplexPk {
+            a: 9,
+            b: 8,
+            c: Some("seven".into()),
+            d: 6,
+            e: 5,
+        };
+        session
+            .query(
+                "INSERT INTO ks.complex_pk (a,b,c,d,e) VALUES (?,?,?,?,?)",
+                input.clone(),
+            )
+            .await
+            .unwrap();
+        let mut rs = session
+            .query(
+                "SELECT * FROM ks.complex_pk WHERE a = 9 and b = 8 and c = 'seven'",
+                &[],
+            )
+            .await
+            .unwrap()
+            .rows
+            .unwrap()
+            .into_typed::<ComplexPk>();
+        let output = rs.next().unwrap().unwrap();
+        assert_eq!(input, output)
+    }
 }
 
 #[tokio::test]
@@ -681,13 +724,13 @@ async fn test_tracing() {
 
 async fn test_tracing_query(session: &Session) {
     // A query without tracing enabled has no tracing uuid in result
-    let untraced_query: Query = Query::new("SELECT * FROM test_tracing_ks.tab".to_string());
+    let untraced_query: Query = Query::new("SELECT * FROM test_tracing_ks.tab");
     let untraced_query_result: QueryResult = session.query(untraced_query, &[]).await.unwrap();
 
     assert!(untraced_query_result.tracing_id.is_none());
 
     // A query with tracing enabled has a tracing uuid in result
-    let mut traced_query: Query = Query::new("SELECT * FROM test_tracing_ks.tab".to_string());
+    let mut traced_query: Query = Query::new("SELECT * FROM test_tracing_ks.tab");
     traced_query.config.tracing = true;
 
     let traced_query_result: QueryResult = session.query(traced_query, &[]).await.unwrap();
@@ -734,7 +777,7 @@ async fn test_tracing_prepare(session: &Session) {
     assert!(untraced_prepared.prepare_tracing_ids.is_empty());
 
     // Preparing a statement with tracing enabled has tracing uuids in result
-    let mut to_prepare_traced = Query::new("SELECT * FROM test_tracing_ks.tab".to_string());
+    let mut to_prepare_traced = Query::new("SELECT * FROM test_tracing_ks.tab");
     to_prepare_traced.config.tracing = true;
 
     let traced_prepared = session.prepare(to_prepare_traced).await.unwrap();
@@ -748,7 +791,7 @@ async fn test_tracing_prepare(session: &Session) {
 
 async fn test_get_tracing_info(session: &Session) {
     // A query with tracing enabled has a tracing uuid in result
-    let mut traced_query: Query = Query::new("SELECT * FROM test_tracing_ks.tab".to_string());
+    let mut traced_query: Query = Query::new("SELECT * FROM test_tracing_ks.tab");
     traced_query.config.tracing = true;
 
     let traced_query_result: QueryResult = session.query(traced_query, &[]).await.unwrap();
@@ -761,7 +804,7 @@ async fn test_get_tracing_info(session: &Session) {
 
 async fn test_tracing_query_iter(session: &Session) {
     // A query without tracing enabled has no tracing ids
-    let untraced_query: Query = Query::new("SELECT * FROM test_tracing_ks.tab".to_string());
+    let untraced_query: Query = Query::new("SELECT * FROM test_tracing_ks.tab");
 
     let mut untraced_row_iter = session.query_iter(untraced_query, &[]).await.unwrap();
     while let Some(_row) = untraced_row_iter.next().await {
@@ -775,7 +818,7 @@ async fn test_tracing_query_iter(session: &Session) {
     assert!(untraced_typed_row_iter.get_tracing_ids().is_empty());
 
     // A query with tracing enabled has a tracing ids in result
-    let mut traced_query: Query = Query::new("SELECT * FROM test_tracing_ks.tab".to_string());
+    let mut traced_query: Query = Query::new("SELECT * FROM test_tracing_ks.tab");
     traced_query.config.tracing = true;
 
     let mut traced_row_iter = session.query_iter(traced_query, &[]).await.unwrap();
@@ -855,9 +898,8 @@ async fn test_tracing_batch(session: &Session) {
 }
 
 async fn assert_in_tracing_table(session: &Session, tracing_uuid: Uuid) {
-    let mut traces_query =
-        Query::new("SELECT * FROM system_traces.sessions WHERE session_id = ?".to_string());
-    traces_query.config.consistency = Consistency::One;
+    let mut traces_query = Query::new("SELECT * FROM system_traces.sessions WHERE session_id = ?");
+    traces_query.config.consistency = Some(Consistency::One);
 
     // Tracing info might not be immediately available
     // If rows are empty perform 8 retries with a 32ms wait in between
@@ -908,3 +950,111 @@ async fn test_await_timed_schema_agreement() {
         .await
         .unwrap();
 }
+
+#[tokio::test]
+async fn test_timestamp() {
+    let uri = std::env::var("SCYLLA_URI").unwrap_or_else(|_| "127.0.0.1:9042".to_string());
+    let session = SessionBuilder::new().known_node(uri).build().await.unwrap();
+
+    session.query("CREATE KEYSPACE IF NOT EXISTS ks WITH REPLICATION = {'class' : 'SimpleStrategy', 'replication_factor' : 1}", &[]).await.unwrap();
+    session
+        .query("DROP TABLE IF EXISTS ks.t_timestamp;", &[])
+        .await
+        .unwrap();
+    session
+        .query(
+            "CREATE TABLE IF NOT EXISTS ks.t_timestamp (a text, b text, primary key (a))",
+            &[],
+        )
+        .await
+        .unwrap();
+
+    session.await_schema_agreement().await.unwrap();
+
+    let query_str = "INSERT INTO ks.t_timestamp (a, b) VALUES (?, ?)";
+
+    // test regular query timestamps
+
+    let mut regular_query = Query::new(query_str.to_string());
+
+    regular_query.set_timestamp(Some(420));
+    session
+        .query(regular_query.clone(), ("regular query", "higher timestamp"))
+        .await
+        .unwrap();
+
+    regular_query.set_timestamp(Some(42));
+    session
+        .query(regular_query.clone(), ("regular query", "lower timestamp"))
+        .await
+        .unwrap();
+
+    // test prepared statement timestamps
+
+    let mut prepared_statement = session.prepare(query_str).await.unwrap();
+
+    prepared_statement.set_timestamp(Some(420));
+    session
+        .execute(&prepared_statement, ("prepared query", "higher timestamp"))
+        .await
+        .unwrap();
+
+    prepared_statement.set_timestamp(Some(42));
+    session
+        .execute(&prepared_statement, ("prepared query", "lower timestamp"))
+        .await
+        .unwrap();
+
+    // test batch statement timestamps
+
+    let mut batch: Batch = Default::default();
+    batch.append_statement(regular_query);
+    batch.append_statement(prepared_statement);
+
+    batch.set_timestamp(Some(420));
+    session
+        .batch(
+            &batch,
+            (
+                ("first query in batch", "higher timestamp"),
+                ("second query in batch", "higher timestamp"),
+            ),
+        )
+        .await
+        .unwrap();
+
+    batch.set_timestamp(Some(42));
+    session
+        .batch(
+            &batch,
+            (
+                ("first query in batch", "lower timestamp"),
+                ("second query in batch", "lower timestamp"),
+            ),
+        )
+        .await
+        .unwrap();
+
+    let mut results = session
+        .query("SELECT a, b, WRITETIME(b) FROM ks.t_timestamp", &[])
+        .await
+        .unwrap()
+        .rows
+        .unwrap()
+        .into_typed::<(String, String, i64)>()
+        .map(Result::unwrap)
+        .collect::<Vec<_>>();
+    results.sort();
+
+    let expected_results = [
+        ("first query in batch", "higher timestamp", 420),
+        ("prepared query", "higher timestamp", 420),
+        ("regular query", "higher timestamp", 420),
+        ("second query in batch", "higher timestamp", 420),
+    ]
+    .iter()
+    .map(|(x, y, t)| (x.to_string(), y.to_string(), *t))
+    .collect::<Vec<_>>();
+
+    assert_eq!(results, expected_results);
+}
